# Upgraded commands extension for hata.
# Work in progress.

from .category import *
from . import checks
from .command import *
from .command_helpers import *
from .command_processor import *
from .content_parser import *
from .context import *
from .cooldown import *
from .exceptions import *
from .responding import *
from .utils import *
from .wrappers import *

__all__ = (
    'checks',
    'cooldown',
    'configure_converter',
    'setup_ext_commands_v2',
    *category.__all__,
    *command.__all__,
    *command_helpers.__all__,
    *command_processor.__all__,
    *content_parser.__all__,
    *context.__all__,
    *cooldown.__all__,
    *exceptions.__all__,
    *responding.__all__,
    *utils.__all__,
    *wrappers.__all__,
)

from .. import register_library_extension, add_library_extension_hook, register_setup_function

from ...discord.client import Client

configure_converter = CommandConverterConfigurerWrapper
cooldown = CommandCooldownWrapper


EXTENSION_SETUP_HOOKS = []

def setup_ext_commands_v2(client, prefix, **kwargs):
    """
    Setups the commands extension of hata on the given client with the given parameters.
    
    Note that this function can be called on a client only once.
    
    The function adds the following attributes to the client:
    - `.command_processor` : ``CommandProcessor``
    - `.commands` : ``_EventHandlerManager``
    
    And the following event handlers are added as well:
    - `message_create` : ``CommandProcessor``
    - `reaction_add` : ``ReactionAddWaitfor`` (Except if other ``EventWaitforBase`` instance is already added.)
    - `reaction_delete` : ``ReactionDeleteWaitfor`` (Except if other ``EventWaitforBase`` instance is already added.)
    
    Parameters
    ----------
    client : ``Client`
        The client on what the extension will be setuped.
    prefix : `str`, (`tuple`, `list`, `deque`) of `str`, or `callable` -> `str`, Optional
        The prefix of the client's command processor.
        
        Can be given as `str`, as `tuple`, `list` or `deque` of `str`, or as a `callable`, what accepts `1` parameter,
        the respective ``Message`` instance and returns `str`.
    **kwargs : Keyword parameters
        Additional keyword parameters to be passed to the created ``CommandProcessor``.
    
    Returns
    -------
    command_processor : ``CommandProcessor``
        The created command processor. Returns `None` if `lite` is given as `True`.
    
    Raises
    ------
    TypeError
        - If `client` was not given as ``Client`` instance.
        - If `prefix` was not given meanwhile `lite` is `False`.
    RuntimeError
        - If the given `client` already has `command_processor` or `commands` attribute.
        - If the given `client` has a ``CommandProcessor`` instance added as `message_create` event,
    """
    if not isinstance(client, Client):
        raise TypeError(f'Expected type `{Client.__name__}` as client, meanwhile got `{client.__class__.__name__}`.')
    
    for attr_name in ('command_processor', 'commands'):
        if hasattr(client, attr_name):
            raise RuntimeError(f'The client already has an attribute named as `{attr_name}`.')
    
    command_processor = CommandProcessor(prefix, **kwargs)
    for hook in EXTENSION_SETUP_HOOKS:
        hook(client, command_processor)
    
    client.events(command_processor)
    
    client.command_processor = command_processor
    client.commands = command_processor.shortcut
    
    return command_processor


def snapshot_hook():
    from . import snapshot


register_library_extension('HuyaneMatsu.commands_v2')
add_library_extension_hook(snapshot_hook, ['HuyaneMatsu.extension_loader'])

def command_utils_hook():
    from . import extension_hook_command_utils

add_library_extension_hook(command_utils_hook, ['HuyaneMatsu.command_utils'])

register_setup_function(
    'HuyaneMatsu.commands_v2',
    setup_ext_commands_v2,
    (
        'prefix',
    ),(
        'precheck',
        'mention_prefix_enabled',
        'category_name_rule',
        'command_name_rule',
        'default_category_name',
        'prefix_ignore_case',
    ),
)



__all__ = ('OpusDecoder', 'OpusEncoder', 'OpusError', )

import ctypes, sys
import ctypes.util
import os.path
from math import log10
from types import ModuleType

opus = None

class OpusError(Exception):
    """
    Exception raised by lib-opus related methods.
    
    Attributes
    ----------
    code : `int`
        Returned error code by lib-opus.
    """
    def __new__(cls, code):
        """
        Raises
        -----
        RuntimeError
            If opus is not loaded.
        """
        if opus is None:
            raise RuntimeError(f'{cls.__name__} cannot be created if opus is not loaded')
        
        return Exception.__new__(cls)
    
    def __init__(self, code):
        """
        Creates an ``OpusError``
        
        Parameters
        ----------
        code : `int`
            Returned error code by lib-opus.
        """
        self.code = code
        msg = opus.opus_strerror(code).decode('utf-8')
        Exception.__init__(self, msg)


OK = 0
ERROR_CODE_BAD_ARG = -1
ERROR_CODE_BUFFER_TOO_SMALL = -2
ERROR_CODE_INTERNAL_ERROR = -3
ERROR_CODE_INVALID_PACKET = -4
ERROR_CODE_UNIMPLEMENTED = -5
ERROR_CODE_INVALID_STATE = -6
ERROR_CODE_ALLOC_FAIL = -7

APPLICATION_AUDIO = 2049
APPLICATION_VOIP = 2048
APPLICATION_LOW_DELAY = 2051

SET_BITRATE = 4002
SET_BANDWIDTH = 4008
SET_INBAND_FEC = 4012
SET_PACKET_LOSS_PERCENTAGE = 4014
SET_SIGNAL = 4024
SET_GAIN = 4034
GET_LAST_PACKET_DURATION = 4039

BANDWIDTH_NARROW = 1101
BANDWIDTH_MEDIUM = 1102
BANDWIDTH_WIDE = 1103
BANDWIDTH_SUPER_WIDE = 1104
BANDWIDTH_FULL = 1105

SIGNAL_TYPE_AUTO = -1000
SIGNAL_TYPE_VOICE = 3001
SIGNAL_TYPE_MUSIC = 3002


def load_opus():
    """
    Loads opus.
    
    Raises
    ------
    Any
    """
    try:
        if sys.platform == 'win32':
            filename = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'bin', f'libopus-0.{"x64" if sys.maxsize>(1<<36) else "x86"}.dll')
        else:
            filename = ctypes.util.find_library('opus')
        
        lib_opus = ctypes.cdll.LoadLibrary(filename)
        
        c_int_ptr = ctypes.POINTER(ctypes.c_int)
        c_int16_ptr = ctypes.POINTER(ctypes.c_int16)
        c_float_ptr = ctypes.POINTER(ctypes.c_float)
        
        EncoderStructPtr=ctypes.POINTER(type('EncoderStruct', (ctypes.Structure,), {},))
        DecoderStructPtr=ctypes.POINTER(type('DecoderStruct', (ctypes.Structure,), {},))
        
        def _err_lt(result, func, args):
            if result < 0:
                raise OpusError(result)
            
            return result
        
        def _err_ne(result, func, args):
            return_code = args[-1]._obj.value
            if return_code < 0:
                raise OpusError(return_code)
            
            return result
        
        functions = []
        
        # register the functions...
        for internal_name, func_name, arg_types, res_type, err_check in (
                ('opus_strerror', 'opus_strerror', (ctypes.c_int,), ctypes.c_char_p, None,),
                ('opus_packet_get_bandwidth', 'opus_packet_get_bandwidth', (ctypes.c_char_p,), ctypes.c_int, _err_lt,),
                ('opus_packet_get_channel_count', 'opus_packet_get_nb_channels', (ctypes.c_char_p,), ctypes.c_int, _err_lt,),
                ('opus_packet_get_frame_count', 'opus_packet_get_nb_frames', (ctypes.c_char_p, ctypes.c_int,), ctypes.c_int, _err_lt,),
                ('opus_packet_get_samples_per_frame', 'opus_packet_get_samples_per_frame', (ctypes.c_char_p, ctypes.c_int,), ctypes.c_int, _err_lt,),
                ('opus_encoder_get_size', 'opus_encoder_get_size', (ctypes.c_int,), ctypes.c_int, None,),
                ('opus_encoder_create', 'opus_encoder_create', (ctypes.c_int, ctypes.c_int, ctypes.c_int, c_int_ptr,), EncoderStructPtr, _err_ne,),
                ('opus_encode', 'opus_encode', (EncoderStructPtr, c_int16_ptr, ctypes.c_int, ctypes.c_char_p, ctypes.c_int32,), ctypes.c_int32, _err_lt,),
                ('opus_encoder_control', 'opus_encoder_ctl', None, ctypes.c_int32, _err_lt,),
                ('opus_encoder_destroy', 'opus_encoder_destroy', (EncoderStructPtr,), None, None,),
                ('opus_decoder_get_size', 'opus_decoder_get_size', (ctypes.c_int,), ctypes.c_int, None,),
                ('opus_decoder_create', 'opus_decoder_create', (ctypes.c_int, ctypes.c_int, c_int_ptr,), DecoderStructPtr, _err_ne,),
                ('opus_decoder_get_sample_count', 'opus_decoder_get_nb_samples', (DecoderStructPtr, ctypes.c_char_p, ctypes.c_int32,), ctypes.c_int, _err_lt,),
                ('opus_decode', 'opus_decode', (DecoderStructPtr, ctypes.c_char_p, ctypes.c_int32, c_int16_ptr, ctypes.c_int, ctypes.c_int,), ctypes.c_int, _err_lt,),
                ('opus_decoder_control', 'opus_decoder_ctl', None, ctypes.c_int32, _err_lt, ),
                ('opus_decoder_destroy', 'opus_decoder_destroy', (DecoderStructPtr,), None, None,),
                    ):
            
            func = getattr(lib_opus, func_name)
            
            try:
                if (arg_types is not None):
                    func.argtypes = arg_types
                
                func.restype = res_type
            except KeyError:
                pass
            
            try:
                if (err_check is not None):
                    func.errcheck = err_check
            except KeyError:
                pass
            
            functions.append((internal_name, func))
    except:
        opus = None
        raise
    else:
        module_path = f'{__name__}.opus'
        try:
            opus = sys.modules[module_path]
        except KeyError:
            opus = sys.modules[module_path] = ModuleType(module_path)
        
        for name, func in functions:
            setattr(opus, name, func)
    finally:
        try:
            actual_module = sys.modules[__name__]
        except KeyError:
            actual_module = sys.modules[__name__] = ModuleType(__name__)
        
        actual_module.opus = opus

try:
    load_opus()
except BaseException:
    pass

SAMPLING_RATE = 48000 #this is the max sadly
CHANNELS = 2
FRAME_LENGTH = 20
SAMPLE_SIZE = 4 # (bit_rate / 8) * CHANNELS (bit_rate == 16)
SAMPLES_PER_FRAME = int(SAMPLING_RATE/1000*FRAME_LENGTH)
FRAME_SIZE = SAMPLES_PER_FRAME*SAMPLE_SIZE
BUFFER_SIZE = 3840
BUFFER_TYPE = ctypes.c_char*BUFFER_SIZE

class OpusEncoder:
    """
    Opus encoder of a ``VoiceClient``.
    
    Attributes
    ----------
    _buffer : `_ctypes.array.c_char_Array_3840`
        A buffer used by the encoder.
    _encoder : `_ctypes.pointer.LP_EncoderStruct`
        Pointer to the C level encoder.
    """
    __slots__ = ('_buffer', '_encoder', )
    def __new__(cls):
        """
        Creates a new opus decoder.
        
        Returns
        -------
        self : ``OpusEncoder``
        
        Raises
        ------
        RuntimeError
            If opus is not loaded.
        """
        if opus is None:
            raise RuntimeError(f'{cls.__name__} cannot be created if opus is not loaded')
        
        encoder = opus.opus_encoder_create(SAMPLING_RATE, CHANNELS, APPLICATION_AUDIO, ctypes.byref(ctypes.c_int()))
        
        self=object.__new__(cls)
        self._encoder = encoder
        self._buffer = BUFFER_TYPE()
        
        opus.opus_encoder_control(encoder, SET_BITRATE, 131072)
        opus.opus_encoder_control(encoder, SET_INBAND_FEC, 1)
        opus.opus_encoder_control(encoder, SET_PACKET_LOSS_PERCENTAGE, 15)
        opus.opus_encoder_control(encoder, SET_BANDWIDTH, BANDWIDTH_FULL)
        opus.opus_encoder_control(encoder, SET_SIGNAL, SIGNAL_TYPE_MUSIC)
        
        return self
    
    def __del__(self):
        """Unallocates `self._encoder`."""
        opus.opus_encoder_destroy(self._encoder)
        self._encoder = None
    
    def set_bitrate(self, kbps):
        """
        Sets the bitrate of the encoder in kilo bit per second. Can be between `16` and `512`.
        
        Parameters
        ----------
        kbps : `int`
        """
        if kbps < 16:
            kbps = 16
        elif kbps > 512:
            kbps = 512
        
        opus.opus_encoder_control(self._encoder, SET_BITRATE, kbps<<10)

    def set_bandwidth(self, bandwidth):
        """
        Set's the band-type of the encoder.
        
        Parameters
        ----------
        bandwidth : `int`
            Requested bandwidth's code.
            
            Possible values
            +-----------------------+-------+
            | Respective name       | Value |
            +=======================+=======+
            | BANDWIDTH_NARROW      | 1101  |
            +-----------------------+-------+
            | BANDWIDTH_MEDIUM      | 1102  |
            +-----------------------+-------+
            | BANDWIDTH_WIDE        | 1103  |
            +-----------------------+-------+
            | BANDWIDTH_SUPER_WIDE  | 1104  |
            +-----------------------+-------+
            | BANDWIDTH_FULL        | 1105  |
            +-----------------------+-------+
        """
        opus.opus_encoder_control(self._encoder, SET_BANDWIDTH, bandwidth)
    
    def set_signal_type(self, signal_type):
        """
        Sets the signal type of the encoder.
        
        Parameters
        ----------
        signal_type : `int`
            Requested signal type.
            
            Possible values
            ---------------
            +-----------------------+-------+
            | Respective name       | Value |
            +=======================+=======+
            | SIGNAL_TYPE_AUTO      | -1000 |
            +-----------------------+-------+
            | SIGNAL_TYPE_VOICE     | 3001  |
            +-----------------------+-------+
            | SIGNAL_TYPE_MUSIC     | 3002  |
            +-----------------------+-------+
        """
        opus.opus_encoder_control(self._encoder, SET_SIGNAL, signal_type)
    
    def set_inband_fec(self, enabled):
        """
        Sets the in-band forward error correction of the encoder.
        
        Parameters
        ----------
        enabled : `bool`
            Whether in-band fec should be enabled.
        """
        opus.opus_encoder_control(self._encoder, SET_INBAND_FEC, int(enabled)) #do we really need to switch from sub-int-type to int-type?
    
    def set_expected_packet_loss_percent(self, percentage):
        """
        Sets the expected packet loss percentage of the encoder.
        
        Parameters
        ----------
        percentage : `float`
            The expected packet loss percentage. Should be between `0` and `100`.
        """
        if percentage < 0.0:
            percentage = 0
        elif percentage > 1.0:
            percentage = 100
        else:
            percentage = int(percentage*100.0)
        
        opus.opus_encoder_control(self._encoder, SET_PACKET_LOSS_PERCENTAGE, percentage)
    
    def encode(self, data):
        """
        Encodes the given `PCM` (Pulse Code Modulation) data.
        
        Parameters
        ----------
        data : `bytes-like`
        
        Returns
        -------
        data : `bytes`
        """
        max_data_bytes = len(data)
        data = ctypes.cast(data, ctypes.POINTER(ctypes.c_int16))
        
        buffer = self._buffer
        end = opus.opus_encode(self._encoder, data, SAMPLES_PER_FRAME, buffer, max_data_bytes)
        return buffer[:end]

class OpusDecoder:
    """
    Opus decoder of a ``VoiceClient``.
    
    Attributes
    ----------
    _buffer : `_ctypes.array.c_char_Array_3840`
        A buffer used by the encoder.
    _decoder : `_ctypes.pointer.LP_DecoderStruct`
        Pointer to the C level decoder.
    """
    __slots__ = ('_buffer', '_decoder',)
    
    def __new__(cls):
        """
        Creates a new opus decoder.
        
        Returns
        -------
        self : ``OpusDecoder``
        
        Raises
        ------
        RuntimeError
            If opus is not loaded.
        """
        if opus is None:
            raise RuntimeError(f'{cls.__name__} cannot be created if opus is not loaded')
        
        decoder = opus.opus_decoder_create(SAMPLING_RATE, CHANNELS, ctypes.byref(ctypes.c_int()))
        
        self = object.__new__(cls)
        self._decoder = decoder
        self._buffer = BUFFER_TYPE()
        
        return self
    
    def __del__(self):
        """Unallocates `self._decoder`."""
        opus.opus_decoder_destroy(self._decoder)
        self._decoder = None
    
    @staticmethod
    def packet_get_frame_count(data):
        """
        Returns the number of frames of the given packet.
        
        Parameters
        ----------
        data : `bytes-like`

        Returns
        -------
        frame_count : `int`
        """
        return opus.opus_packet_get_frame_count(data, len(data))
    
    @staticmethod
    def packet_get_channel_count(data):
        """
        Returns the number of channels of the given packet.
        
        Parameters
        ----------
        data : `bytes-like`

        Returns
        -------
        channel_count : `int`
        """
        return opus.opus_packet_get_channel_count(data) #number of channels

    @staticmethod
    def packet_get_samples_per_frame(data):
        """
        Returns the number of samples per frame for the given data.
        
        Parameters
        ----------
        data : `bytes-like`
        
        Returns
        -------
        samples_per_frame : `int`
        """
        return opus.opus_packet_get_samples_per_frame(data, SAMPLING_RATE)
    
    def set_gain(self, adjustment): #sets decibel
        """
        Sets the gain of the decoder in decibel
        
        Parameters
        ----------
        adjustment : `float`
        """
        adjustment = int(256*adjustment)
        if adjustment > 32767:
            adjustment = 32767
        elif adjustment < -32768:
            adjustment = -32768
        
        opus.opus_decoder_control(self._decoder, SET_GAIN, adjustment)
    
    def set_volume_percent(self, percent):
        """
        Sets the output's volume percentage of the decoder.
        
        Parameters
        ----------
        percent : `float`
        """
        return self.set_gain(20.0*log10(percent)) # amplitude ratio
    
    def _get_last_packet_duration(self):
        """
        Returns the last packet's duration
        
        Returns
        -------
        duration : `int`
        """
        obj = ctypes.c_int32()
        opus.opus_decoder_control(self._decoder, GET_LAST_PACKET_DURATION, ctypes.byref(obj))
        return obj.value
    
    def decode(self, data):
        """
        Decodes the given `PCM` (Pulse Code Modulation) data.
        
        Parameters
        ----------
        data : `bytes-like`
        
        Returns
        -------
        data : `bytes`
        """
        frame_size = self.packet_get_frame_count(data)*self.packet_get_samples_per_frame(data)
        
        buffer = self._buffer
        buffer_ptr = ctypes.cast(buffer, ctypes.POINTER(ctypes.c_int16))
        
        end = opus.opus_decode(self._decoder, data, len(data), buffer_ptr, frame_size, True)
        return bytes(buffer[:((end<<1)*CHANNELS)])



__all__ = ('Invite',)

from datetime import datetime

from ..bases import DiscordEntity, instance_or_id_to_instance
from ..preconverters import preconvert_str, preconvert_int, preconvert_bool, preconvert_preinstanced_type
from ..utils import timestamp_to_datetime, DISCORD_EPOCH_START
from ..core import GUILDS, CHANNELS, INVITES
from ..user import User, ZEROUSER, ClientUserBase
from ..guild import create_partial_guild_from_data, Guild
from ..channel import create_partial_channel_from_data, ChannelText, ChannelGroup, ChannelVoice, ChannelStore, \
    ChannelDirectory
from ..application import Application
from ..http import urls as module_urls

from .preinstanced import InviteTargetType
from .invite_stage import InviteStage

# Experimental addition

EMBEDDED_ACTIVITY_NAME_TO_APPLICATION_ID = {
    'Betrayal.io': 773336526917861400,
    'Chess in The Park': 832012774040141894,
    'Fishington.io': 814288819477020702,
    'Poker Night': 755827207812677713,
    'YouTube Together': 755600276941176913,
}

EMBEDDED_ACTIVITY_APPLICATION_ID_TO_NAME = {
    value: key for key, value in EMBEDDED_ACTIVITY_NAME_TO_APPLICATION_ID.items()
}

class Invite(DiscordEntity, immortal=True):
    """
    Represents a Discord Invite.
    
    Attributes
    ---------
    approximate_online_count : `int`
        The approximate amount of online users at the respective guild (or group channel). If not included, then set
        as `0`.
    approximate_user_count : `int`
        The approximate amount of users at the respective guild (or group channel). If not included, then set as `0`.
    channel : `None`, ``ChannelText``, ``ChannelVoice``, ``ChannelStore``, ``ChannelGroup`` or ``ChannelDirectory``
        The channel where the invite redirects. If it is announcements or store channel, then the invite is a lurk
        invite. If channel data was not sent with the invite's, then this attribute is set as `None`.
    code : `str`
        The invite's unique identifier.
    created_at : `datetime`
        When the invite was created. Defaults to Discord epoch.
    guild : `None` or ``Guild``
        The guild the invite is for. If not included or if the invite's channel is a group channel, then set as
        `None`.
    inviter : ``ClientUserBase``
        The creator of the invite. If not included, then set as `ZEROUSER`.
    max_age : `None` or `int`
        The time in seconds after the invite will expire. If not included, then set as `None`.
        
        If the invite was created with max age as `0`, then this value will be negative instead of the expected `0`.
    max_uses : `None` or `int`
        How much times the invite can be used. If not included, then set as `None`.
        
        If the invite has no use limit, then this value is set as `0`.
    partial : `bool`
        Whether the invite is only partially loaded.
    stage : `None` or ``InviteStage``
        A invite stage instance representing the stage to which the invite is created to.
    target_type : ``InviteTargetType``
        The invite's target type.
    target_application : `None` or ``Application``
        The invite's target application.
    target_user : ``ClientUserBase``
        The target of the invite if applicable. Defaults to `ZEROUSER`.
    temporary : `bool`
        Whether this invite only grants temporary membership.
        
        When the user goes offline, they get kicked, except if they got a role meanwhile.
    uses : `None` or `int`
        The amount how much times the invite was used. If not included, set as `None`.
    """
    __slots__ = ('approximate_user_count', 'approximate_online_count', 'channel', 'code', 'created_at', 'guild',
        'inviter', 'max_age', 'max_uses', 'partial', 'target_application', 'stage', 'target_type', 'target_user',
        'temporary', 'uses',)
    
    def __new__(cls, data, data_partial):
        """
        Creates an invite from the given invite data. The invite data can be requested or received through the gateway
        as well.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Invite data.
        """
        code = data['code']
        
        try:
            self = INVITES[code]
        except KeyError:
            self = object.__new__(cls)
            self.code = code
            self.partial = data_partial
            INVITES[code] = self
            updater = cls._set_attributes
        else:
            if self.partial:
                if data_partial:
                    updater = cls._update_attributes
                else:
                    updater = cls._set_attributes
                    self.partial = False
            else:
                if data_partial:
                    updater = cls._update_counts_only
                else:
                    updater = cls._update_attributes
        
        updater(self, data)
        return self
    
    @classmethod
    def _create_vanity(cls, guild, data):
        """
        Creates a vanity invite from the given guild and from the given invite data.
        
        Parameters
        ----------
        guild : ``Guild``
            The respective guild of the vanity invite.
        data : `dict` of (`str`, `Any`) items
            Invite data requested from Discord.
        
        Returns
        -------
        invite : ``Invite``
        """
        code = guild.vanity_code
        try:
            self = INVITES[code]
        except KeyError:
            self = object.__new__(cls)
        
        self.code = code
        self.inviter = ZEROUSER
        self.uses = None
        self.max_age = None
        self.max_uses = None
        self.temporary = False
        self.created_at = DISCORD_EPOCH_START
        self.guild = guild
        try:
            channel_data = data['channel']
        except KeyError:
            channel = None
        else:
            channel = create_partial_channel_from_data(channel_data, guild.id)
        self.channel = channel
        self.approximate_online_count = 0
        self.approximate_user_count = 0
        self.target_type = InviteTargetType.none
        self.target_user = ZEROUSER
        self.target_application = None
        self.partial = True
        self.stage = None
        
        return self
    
    def __repr__(self):
        """Returns the representation of the invite."""
        return f'<{self.__class__.__name__} code={self.code!r}>'
    
    def __hash__(self):
        """Returns the invite's code's hash."""
        return hash(self.code)
    
    url = property(module_urls.invite_url)
    
    @property
    def id(self):
        """
        Compatibility property with other Discord entities.
        
        Returns
        -------
        id : `int` = `0`
        """
        return 0
    
    # When we update it we get only a partial invite from Discord. So sad.
    def _set_attributes(self, data):
        """
        Updates the invite by overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Invite data.
        """
        try:
            guild_data = data['guild']
        except KeyError:
            try:
                guild_id = data['guild_id']
            except KeyError:
                guild = None
            else:
                guild_id = int(guild_id)
                guild = GUILDS.get(guild_id, None)
        else:
            guild = create_partial_guild_from_data(guild_data)
        
        self.guild = guild
        
        try:
            channel_data = data['channel']
        except KeyError:
            try:
                channel_id = data['channel_id']
            except KeyError:
                channel = None
            else:
                channel_id = int(channel_id)
                channel = CHANNELS.get(channel_id, None)
        else:
            channel = create_partial_channel_from_data(channel_data, guild.id)
        self.channel = channel
        
        try:
            approximate_online_count = data['approximate_presence_count']
        except KeyError:
            approximate_online_count = 0
        else:
            if (guild is not None):
                guild.approximate_online_count = approximate_online_count
        
        self.approximate_online_count = approximate_online_count
        
        try:
            approximate_user_count = data['approximate_member_count']
        except KeyError:
            approximate_user_count = 0
        else:
            if (guild is not None):
                guild.approximate_user_count = approximate_user_count
            
        self.approximate_user_count = approximate_user_count
        
        try:
            inviter_data = data['inviter']
        except KeyError:
            inviter = ZEROUSER
        else:
            inviter = User(inviter_data)
        self.inviter = inviter
        
        self.uses = data.get('uses', None)
        self.max_age = data.get('max_age', None)
        self.max_uses = data.get('max_uses', None)
        self.temporary = data.get('temporary', True)
        
        try:
            created_at_data = data['created_at']
        except KeyError:
            created_at = DISCORD_EPOCH_START
        else:
            created_at = timestamp_to_datetime(created_at_data)
        self.created_at = created_at
        
        self.target_type = InviteTargetType.get(data.get('target_type', 0))
        
        try:
            target_user_data = data['target_user']
        except KeyError:
            target_user = ZEROUSER
        else:
            target_user = User(target_user_data)
        
        self.target_user = target_user
        
        try:
            target_application_data = data['target_application']
        except KeyError:
            target_application = None
        else:
            target_application = Application(target_application_data)
        
        self.target_application = target_application
        
        try:
            invite_stage_data = data['stage_instance']
        except KeyError:
            invite_stage = None
        else:
            invite_stage = InviteStage(invite_stage_data, guild)
        self.stage = invite_stage
    
    
    def _update_attributes(self, data):
        """
        Updates the invite with the given data. Only updates the attributes, which's respective value is included.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Invite data.
        """
        guild = self.guild
        if guild is None:
            try:
                guild_data = data['guild']
            except KeyError:
                try:
                    guild_id = data['guild_id']
                except KeyError:
                    guild = None
                else:
                    guild_id = int(guild_id)
                    guild = GUILDS.get(guild_id, None)
            else:
                guild = create_partial_guild_from_data(guild_data)
            
            self.guild = guild
        
        if self.channel is None:
            try:
                channel_data = data['channel']
            except KeyError:
                try:
                    channel_id = data['channel_id']
                except KeyError:
                    channel = None
                else:
                    channel_id = int(channel_id)
                    channel = CHANNELS.get(channel_id, None)
            else:
                channel = create_partial_channel_from_data(channel_data, guild.id)
            
            self.channel = channel
        
        self._update_counts_only(data)
        
        try:
            inviter_data = data['inviter']
        except KeyError:
            pass
        else:
            self.inviter = User(inviter_data)
        
        try:
            self.uses = data['uses']
        except KeyError:
            pass
        
        try:
            self.max_age = data['max_age']
        except KeyError:
            pass
        
        try:
            self.max_uses = data['max_uses']
        except KeyError:
            pass
        
        try:
            self.temporary = data['temporary']
        except KeyError:
            pass
        
        try:
            created_at_data = data['created_at']
        except KeyError:
            pass
        else:
            self.created_at = timestamp_to_datetime(created_at_data)
        
        try:
            target_type_value = data['target_type']
        except KeyError:
            pass
        else:
            self.target_type = InviteTargetType.get(target_type_value)
        
        try:
            target_user_data = data['target_user']
        except KeyError:
            pass
        else:
            self.target_user = User(target_user_data)
    
        try:
            target_application_data = data['target_application']
        except KeyError:
            pass
        else:
            self.target_application = Application(target_application_data)
        
        try:
            invite_stage_data = data['stage_instance']
        except KeyError:
            pass
        else:
            self.stage = InviteStage(invite_stage_data, guild)
    
    
    def _update_counts_only(self, data):
        """
        Updates the invite's counts if given.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received invite data.
        """
        guild = self.guild
        try:
            approximate_online_count = data['approximate_presence_count']
        except KeyError:
            pass
        else:
            self.approximate_online_count = approximate_online_count
            if (guild is not None):
                guild.approximate_online_count = approximate_online_count
        
        try:
            approximate_user_count = data['approximate_member_count']
        except KeyError:
            pass
        else:
            self.approximate_user_count = approximate_user_count
            if (guild is not None):
                guild.approximate_user_count = approximate_user_count
    
    
    @classmethod
    def precreate(cls, code, **kwargs):
        """
        Precreates an invite object with the given parameters.
        
        Parameters
        ----------
        code : `str`
            The invite's code.
        **kwargs : keyword parameters
            Additional predefined attributes for the invite.
        
        Other Parameters
        ----------------
        channel : `None`, ``ChannelText``, ``ChannelVoice``, ``ChannelStore``, ``ChannelGroup`` or
                ``ChannelDirectory``, Optional (Keyword only)
            The channel where the invite redirects.
        created_at : `datetime`, Optional (Keyword only)
            When the invite was created.
        guild : `None` or ``Guild``, Optional (Keyword only)
            The guild the invite is for.
        inviter : `int`, `str`, ``ClientUserBase``, Optional (Keyword only)
            The creator of the invite.
        max_age : `None` or `int`, Optional (Keyword only)
            The time in seconds after the invite will expire.
        max_uses : `None` or `int`, Optional (Keyword only)
            How much times the invite can be used.
        approximate_online_count : `int`, Optional (Keyword only)
            The amount of online users at the respective guild (or group channel).
        target_type : `int` or ``InviteTargetType``, Optional (Keyword only)
            The invite's target type.
        target_application : `int`, `str` or ``Application``, Optional (Keyword only)
            The target application of the invite.
        target_user : `int`, `str`, ``ClientUserBase``, Optional (Keyword only)
            The target user of the invite.
        temporary : `bool`, Optional (Keyword only)
            Whether this invite only grants temporary membership.
        approximate_user_count : `int`, Optional (Keyword only)
            The amount of users at the respective guild (or group channel).
        uses : `None` or `int`, Optional (Keyword only)
            The amount how much times the invite was used.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        code = preconvert_str(code, 'code', 2, 32)

        if kwargs:
            processable = []
            for key in ('uses', 'max_age', 'max_uses' ):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    pass
                else:
                    if (value is not None):
                        value = preconvert_int(value, key, 0, (1<<64)-1)
                        processable.append((key, value))
            
            try:
                temporary = kwargs.pop('temporary')
            except KeyError:
                pass
            else:
                temporary = preconvert_bool(temporary, 'temporary')
                processable.append(('temporary', temporary))
            
            try:
                created_at = kwargs.pop('created_at')
            except KeyError:
                pass
            else:
                created_at_type = created_at.__class__
                if (created_at_type is not datetime):
                    raise TypeError(f'`\'created_at\'` can be `int` instance, got {created_at_type.__name__}.')
                processable.append(('created_at', created_at))
            
            for key in ('inviter', 'target_user',):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    pass
                else:
                    value = instance_or_id_to_instance(value, ClientUserBase, key)
                    processable.append((key, value))
            
            for key in ('approximate_online_count', 'approximate_user_count',):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    pass
                else:
                    value = preconvert_int(value, key, 0, (1<<64)-1)
                    processable.append((key, value))
            
            for key, type_ in (
                    ('guild', Guild),
                    ('channel', (ChannelText, ChannelGroup, ChannelVoice, ChannelStore, ChannelDirectory)),
                        ):
                
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    pass
                else:
                    if (value is not None):
                        value = instance_or_id_to_instance(value, type_, key)
                        processable.append((key, value))
                    
            try:
                target_type = kwargs.pop('target_type')
            except KeyError:
                pass
            else:
                target_type = preconvert_preinstanced_type(target_type, InviteTargetType, 'target_type')
                processable.append(('target_type', target_type))
            
            try:
                target_application = kwargs.pop('target_application')
            except KeyError:
                pass
            else:
                value = instance_or_id_to_instance(target_application, Application, key)
                processable.append(('target_application', value))
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}.')
        
        else:
            processable = None
        
        try:
            self = INVITES[code]
        except KeyError:
            self = object.__new__(cls)
            self.code = code
            self.inviter = ZEROUSER
            self.uses = None
            self.max_age = None
            self.max_uses = None
            self.temporary = False
            self.created_at = DISCORD_EPOCH_START
            self.guild = None
            self.channel = None
            self.approximate_online_count = 0
            self.approximate_user_count = 0
            self.target_type = InviteTargetType.none
            self.target_user = None
            self.target_user = ZEROUSER
            self.partial = True
            self.stage = None
            
            INVITES[code] = self
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self



__all__ = ()

import ssl, selectors
from collections import deque
import socket as module_socket

from .futures import Future

if hasattr(module_socket, 'TCP_NODELAY'):
    def _set_nodelay(socket):
        if (socket.family in (module_socket.AF_INET, module_socket.AF_INET6) and
            socket.type == module_socket.SOCK_STREAM and socket.proto == module_socket.IPPROTO_TCP):
            socket.setsockopt(module_socket.IPPROTO_TCP, module_socket.TCP_NODELAY, 1)
else:
    def _set_nodelay(socket):
        pass

MAX_SIZE = 262144

UNWRAPPED = 'UNWRAPPED'
DO_HANDSHAKE = 'DO_HANDSHAKE'
WRAPPED = 'WRAPPED'
SHUTDOWN = 'SHUTDOWN'

class _SSLPipe:
    """
    An SSL pipe.
    
    Allows you to communicate with an SSL/TLS protocol instance through memory buffers. It can be used to implement a
    security layer for an existing connection where you don't have access to the connection's file descriptor, or for
    some reason you don't want to use it.
    
    An ``_SSLPipe`` can be in `wrapped` and `unwrapped` mode. In unwrapped mode, data is passed through untransformed.
    In wrapped mode, application level data is encrypted to SSL record level data and vice versa. The SSL record level
    is the lowest level in the SSL protocol suite and is what travels as-is over the wire.
    
    An ``_SSLPipe`` initially is in `unwrapped` mode. To start SSL, call ``.do_handshake``, to shutdown SSL, call
    ``.unwrap``.
    
    Attributes
    ----------
    _handshake_cb : `None` or `callable`
        A Callback which will be called when handshake is completed. Set by ``.do_handshake``.
        
        Should accept the following parameters:
        +-----------------------+---------------------------+
        | Respective name       | Value                     |
        +=======================+===========================+
        | handshake_exception   | `None` or `BaseException` |
        +-----------------------+---------------------------+
        
        If the handshake is successful, then the `handshake_exception` is given as `None`, else as an exception
        instance.
    _incoming : `ssl.MemoryBIO`
        Does the incoming data encryption/decryption.
    _outgoing : `ssl.MemoryBIO`
        Does the outgoing data encryption/decryption.
    _shutdown_cb : `None` or `callable`
        A callback which will be called when the shutdown is completed. Set by ``.shutdown``.
        
        Should accept no parameters.
    context : `ssl.SSLContext`
        The SSL pipe's SSL type.
    need_ssl_data : `bool`
        Whether more record level data is needed to complete a handshake that is currently in progress.
    server_hostname : `None` or `str`
        The ssl protocol's server hostname if applicable.
    server_side : `bool`
        Whether the ssl protocol is server side.
    ssl_object : `None` or `ssl.SSLObject`
        SSL object connecting ``._incoming`` and ``._outgoing`` memory bios set at ``.do_handshake``.
    state : `str`
        The state of the ``_SSLPipe``.
        
        Can be set as one of:
        
        +-------------------+-------------------+
        | Respective name   | Value             |
        +===================+===================+
        | UNWRAPPED         | `'UNWRAPPED'`     |
        +-------------------+-------------------+
        | DO_HANDSHAKE      | `'DO_HANDSHAKE'`  |
        +-------------------+-------------------+
        | WRAPPED           | `'WRAPPED'`       |
        +-------------------+-------------------+
        | SHUTDOWN          | `'SHUTDOWN'`      |
        +-------------------+-------------------+
        
        Note, that ``.state`` is checked by memory address and not by value.
    """
    __slots__ = ('_handshake_cb', '_incoming', '_outgoing', '_shutdown_cb', 'context', 'need_ssl_data',
        'server_hostname', 'server_side', 'ssl_object', 'state',)
    
    def __init__(self, context, server_side, server_hostname):
        """
        Creates a new ``_SSLPipe`` instance with the given parameters.
        
        Parameters
        ----------
        context : `ssl.SSLContext`
            The SSL pipe's SSL type.
        server_side : `bool`
            Whether the ssl protocol is server side.
        server_hostname : `None` or `str`
            The ssl protocol's server hostname if applicable.

        """
        self.context = context
        self.server_side = server_side
        self.server_hostname = server_hostname
        self.state = UNWRAPPED
        self._incoming = ssl.MemoryBIO()
        self._outgoing = ssl.MemoryBIO()
        self.ssl_object = None
        self.need_ssl_data = False
        self._handshake_cb = None
        self._shutdown_cb = None
    
    
    def is_wrapped(self):
        """
        Returns whether a security layer is currently is effect.
        
        In handshake or on in shutdown is always `False`.
        
        Returns
        -------
        is_wrapped : `bool`
        """
        return (self.state is WRAPPED)
    
    def do_handshake(self, callback=None):
        """
        Starts the SSL handshake.
        
        Parameters
        ----------
        callback : `None` or `callable`, Optional
            A Callback which will be called when handshake is completed.
            
            Should accept the following parameters:
            +-------------------+---------------------------+
            | Respective name   | Value                     |
            +===================+===========================+
            | handshake_exception     | `None` or `BaseException` |
            +-------------------+---------------------------+
            
            If the handshake is successful, then the `handshake_exception` is given as `None`, else as an exception
            instance.
            
        Returns
        -------
        ssl_data : `list` of `bytes`
            A list of SSL data. Always an empty list is returned.
        
        Raises
        ------
        RuntimeError
            If the handshake is in progress or is already completed.
        """
        if self.state is not UNWRAPPED:
            raise RuntimeError('Handshake in progress or completed.')
        
        self.ssl_object = self.context.wrap_bio(self._incoming, self._outgoing, server_side=self.server_side,
            server_hostname=self.server_hostname)
        
        self.state = DO_HANDSHAKE
        self._handshake_cb = callback
        ssl_data, application_data = self.feed_ssl_data(b'', only_handshake=True)
        
        return ssl_data
    
    def shutdown(self, callback=None):
        """
        Starts the SSL shutdown process.
        
        Parameters
        ----------
        callback : `None` or `callable`, Optional
            A callback which will be called when the shutdown is completed.
            
            Should accept no parameters.
        
        Returns
        -------
        ssl_data : `list` of `bytes`
            A list of SSL data.
        
        Raises
        ------
        RuntimeError
            - If the security layer is not yet present.
            - If shutdown is already in progress.
        """
        state = self.state
        if state is UNWRAPPED:
            raise RuntimeError('No security layer present.')
        if state is SHUTDOWN:
            raise RuntimeError('Shutdown in progress.')
        
        self.state = SHUTDOWN
        self._shutdown_cb = callback
        ssl_data, application_data = self.feed_ssl_data(b'')
        return ssl_data
    
    def feed_eof(self):
        """
        Sends an eof.
        
        Raises
        ------
        ssl.SSLError
            If the eof was unexpected. The raised ssl error has `errno` set as `ssl.SSL_ERROR_EOF`.
        """
        self._incoming.write_eof()
        # Ignore return.
        self.feed_ssl_data(b'')
    
    def feed_ssl_data(self, data, only_handshake=False):
        """
        Feed SSL record level data into the pipe.
        
        Parameters
        ----------
        data : `bytes`
            If empty `bytes` is given, then it will get ssl_data for the handshake initialization.
        
        Returns
        -------
        ssl_data : `list` of `bytes`
            Contains the SSL data that needs to be sent to the remote SSL.
        application_data : `list` of `bytes`
            Plaintext data that needs to be forwarded to the application.
            
            Might contain an empty `bytes`, what indicates that ``.shutdown`` should be called.
        """
        state = self.state
        ssl_data = []
        application_data = []
        
        if state is UNWRAPPED:
            # If unwrapped, pass plaintext data straight through.
            if data:
                application_data.append(data)
            
            return ssl_data, application_data
        
        self.need_ssl_data = False
        if data:
            self._incoming.write(data)
        
        try:
            if state is DO_HANDSHAKE:
                # Call do_handshake() until it doesn't raise anymore.
                self.ssl_object.do_handshake()
                self.state = state = WRAPPED
                
                handshake_cb = self._handshake_cb
                if (handshake_cb is not None):
                    handshake_cb(None)
                
                if only_handshake:
                    return ssl_data, application_data
                # Handshake done: execute the wrapped block
            
            if state is WRAPPED:
                # Main state: read data from SSL until close_notify.
                while True:
                    chunk = self.ssl_object.read(MAX_SIZE)
                    application_data.append(chunk)
                    if not chunk:  # close_notify
                        break
            
            elif state is SHUTDOWN:
                # Call shutdown() until it doesn't raise anymore.
                self.ssl_object.unwrap()
                self.ssl_object = None
                self.state = UNWRAPPED
                
                shutdown_cb = self._shutdown_cb
                if (shutdown_cb is not None):
                    shutdown_cb()
            
            elif state is UNWRAPPED:
                # Drain possible plaintext data after close_notify.
                application_data.append(self._incoming.read())
        
        except (ssl.SSLError, ssl.CertificateError) as err:
            err_number = getattr(err, 'errno', None)
            if err_number not in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE, ssl.SSL_ERROR_SYSCALL):
                if self.state is DO_HANDSHAKE:
                    handshake_cb = self._handshake_cb
                    if (handshake_cb is not None):
                        handshake_cb(err)
                raise
            
            self.need_ssl_data = (err_number == ssl.SSL_ERROR_WANT_READ)

        # Check for record level data that needs to be sent back. Happens for the initial handshake and renegotiation.
        if self._outgoing.pending:
            ssl_data.append(self._outgoing.read())
        
        return ssl_data, application_data

    def feed_application_data(self, data, offset):
        """
        Feed plaintext data into the pipe.
        
        If the returned `offset` is less than the returned data's total length in bytes, that case is called
        `short write`. If `short-write` happens, than the same `data` should be passed to ``.feed_application_data``
        next time. (This is an OpenSSL requirement.) A further particularity is that a short write will always have
        `offset = 0`, because the `_ssl` module does not enable partial writes. And even though the offset is `0`,
        there will still be encrypted data in ssl_data.
        
        Returns
        -------
        ssl_data : `list` of `bytes-like`
            Containing record level data that needs to be sent to the remote SSL instance.
        offset : `int`
            The number of the processed plaintext data in bytes. Might be less than the total length of the returned
            data.
        
        Raises
        ------
        ssl.SSLError
            Any unexpected SSL error.
        """
        ssl_data = []
        if self.state is UNWRAPPED:
            # pass through data in unwrapped mode
            if offset < len(data):
                ssl_data.append(memoryview(data)[offset:])
            offset = len(data)
        else:
            view = memoryview(data)
            while True:
                self.need_ssl_data = False
                try:
                    if offset < len(view):
                        offset += self.ssl_object.write(view[offset:])
                except ssl.SSLError as err:
                    # It is not allowed to call `.write` after ``.unwrap`` until the close_notify is acknowledged. We
                    # return the condition to the caller as a short write.
                    if err.reason == 'PROTOCOL_IS_SHUTDOWN':
                        err.errno = ssl.SSL_ERROR_WANT_READ
                        need_ssl_data = True
                    else:
                        errno = getattr(err, 'errno', None)
                        if errno is None:
                            raise
                        
                        if errno == ssl.SSL_ERROR_WANT_READ:
                            need_ssl_data = True
                        elif errno == ssl.SSL_ERROR_WANT_WRITE or errno == ssl.SSL_ERROR_SYSCALL:
                            need_ssl_data = False
                        else:
                            raise
                    
                    self.need_ssl_data = need_ssl_data
    
                # See if there's any record level data back for us.
                if self._outgoing.pending:
                    ssl_data.append(self._outgoing.read())
                
                if offset == len(view) or self.need_ssl_data:
                    break
        
        return ssl_data, offset

class _SSLProtocolTransport:
    """
    Asynchronous ssl protocol's read-write  transport implementation.
    
    Attributes
    ----------
    app_protocol : `Any`
        Asynchronous protocol implementation.
    closing : `bool`
        Whether the ssl protocol transport is closing.
    ssl_protocol : ``SSLProtocol``
        The respective ssl protocol, what's `.app_protocol` is the ``_SSLProtocolTransport`` instance.
    """
    __slots__ = ('app_protocol', 'closing', 'ssl_protocol',)
    def __init__(self, ssl_protocol, app_protocol):
        """
        Creates a  new ``_SSLProtocolTransport`` instance.
        
        Parameters
        ----------
        ssl_protocol : ``SSLProtocol``
            The respective ssl protocol, what's `.app_protocol` is the ``_SSLProtocolTransport`` instance.
        app_protocol : `Any`
            Asynchronous protocol implementation.
        """
        self.ssl_protocol = ssl_protocol
        self.app_protocol = app_protocol
        self.closing = False
    
    def __repr__(self):
        """Returns the ssl protocol transport's representation."""
        result = [
            '<',
            self.__class__.__name__,
            ' ',
                ]
        
        if self.closing:
            result.append('closing')
            add_comma = True
        else:
            add_comma = False
        
        ssl_protocol = self.ssl_protocol
        if (ssl_protocol is not None):
            if add_comma:
                result.append(', ')
            else:
                add_comma = True
            
            result.append('ssl_protocol=')
            result.append(ssl_protocol.__class__.__name__)
        
        app_protocol = self.ssl_protocol
        if (app_protocol is not None):
            if add_comma:
                result.append(', ')
            
            result.append('app_protocol=')
            result.append(app_protocol.__class__.__name__)
        
        result.append('>')
        
        return ''.join(result)
    
    def __del__(self):
        """
        Closes the ssl protocol transport if not yet closed.
        """
        self.close()
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self.ssl_protocol.get_extra_info(name, default)
    
    def get_protocol(self):
        """
        Gets the ssl protocol transport's actual protocol.
        
        Returns
        -------
        protocol : `Any`
            Asynchronous protocol implementation.
        """
        return self.app_protocol
    
    def set_protocol(self, protocol):
        """
        Sets a new protocol to the transport.
        
        Parameters
        ----------
        protocol : `Any`
            Asynchronous protocol implementation.
        """
        self.app_protocol = protocol
        self.ssl_protocol.app_protocol = protocol
    
    def is_closing(self):
        """
        Returns whether the ssl protocol transport is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self.closing
    
    def close(self):
        """
        Starts the shutdown process of the ssl protocol transport.
        """
        if self.closing:
            return
        
        self.closing = True
        self.ssl_protocol._start_shutdown()
    
    def abort(self):
        """
        Close the transport immediately.
        
        Buffered data will be lost and no more data will be received.
        
        The respective protocol's `.connection_lost` will be eventually called with `None`.
        """
        self.ssl_protocol._abort()
    
    def write(self, data):
        """
        Write the given `data` to the transport.
        
        Do not blocks, but queues up the data instead to be sent as can.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to send.
        
        Raises
        ------
        TypeError
            If `data` was not given as `bytes-like`.
        """
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(f'`data` expecting a `bytes-like`, got {data.__class__.__name__}.')
        
        if data:
            self.ssl_protocol._write_application_data(data)
    
    def writelines(self, lines):
        """
        Writes an `iterable` of `bytes-like` to the transport.
        
        Parameters
        ----------
        lines : `iterable` of `bytes-like`
            An iterable of data to send.
        
        Raises
        ------
        TypeError
            If `list_of_data` was not given as `iterable` of `bytes-like`.
        """
        self.write(b''.join(lines))
    
    def write_eof(self):
        """
        Writes eof to the transport if it supports it.
        """
        pass
    
    def can_write_eof(self):
        """
        Return whether the transport supports ``.write_eof``.
        
        Returns
        -------
        can_write_eof : `bool`
            ``_SSLProtocolTransport`` instances always return `False`.
        """
        return False
    
    def pause_reading(self):
        """
        Pauses the receiving end.
        
        No data will be passed to the respective protocol's ``.data_received`` method until ``.resume_reading`` is
        called.
        
        Returns
        -------
        reading_paused : `bool`
            Whether reading was paused.
        """
        transport = self.ssl_protocol.transport
        if transport is None:
            return False
        
        return transport.pause_reading()
    
    def resume_reading(self):
        """
        Resumes the receiving end.
        
        Data received will once again be passed to the respective protocol's ``.data_received`` method.
        
        Returns
        -------
        reading_resume : `bool`
            Whether reading was resumed.
        """
        transport = self.ssl_protocol.transport
        if transport is None:
            return False
            
        transport.resume_reading()
        return True
    
    def get_write_buffer_size(self):
        """
        Return the current size of the write buffer.
        
        Returns
        -------
        get_write_buffer_size : `int`
        """
        transport = self.ssl_protocol.transport
        if transport is None:
            write_buffer_size = 0
        else:
            write_buffer_size = transport.get_write_buffer_size()
        
        return write_buffer_size
    
    def get_write_buffer_limits(self):
        """
        Returns the low and the high water of the transport.
        
        Returns
        -------
        low_water : `int`
            The ``.ssl_protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.
        high_water : `int`
            The ``.ssl_protocol`` is paused writing when the buffer size passes the high water mark. Defaults to
            `65536`.
        """
        transport = self.ssl_protocol.transport
        if transport is None:
            return 16384, 65536
        
        return transport.get_write_buffer_limits()
    
    def set_write_buffer_limits(self, low=None, high=None):
        """
        Set the high- and low-water limits for write flow control.
        
        These two values control when to call the protocol's ``.pause_writing`` and ``.resume_writing`` methods. If
        specified, the low-water limit must be less than or equal to the high-water limit. Neither value can be
        negative. The defaults are implementation-specific. If only the high-water limit is given, the low-water limit
        defaults to an implementation-specific value less than or equal to the high-water limit. Setting high to zero
        forces low to zero as well, and causes ``.pause_writing`` to be called whenever the buffer becomes non-empty.
        Setting low to zero causes ``.resume_writing`` to be called only once the buffer is empty. Use of zero for
        either limit is generally sub-optimal as it reduces opportunities for doing I/O and computation concurrently.
        
        Parameters
        ----------
        high : `None` or `int`, Optional
            High limit to stop reading if reached.
        low : `None` or `int`, Optional
            Low limit to start reading if reached.
        
        Raises
        ------
        ValueError
            If `low` is lower than `0` or if `low` is higher than `high`.
        """
        transport = self.ssl_protocol.transport
        if (transport is not None):
            transport.set_write_buffer_limits(high, low)


class SSLProtocol:
    """
    Asynchronous SSL protocol implementation on top of a `socket`. Uses `ssl.MemoryBIO` instances for incoming and
    outgoing data.
    
    Attributes
    ----------
    _call_connection_made : `bool`
        Whether the the ``.app_protocol``'s `.connection_made` should be called when handshake is completed.
    _connection_made_waiter : `None` or ``Future``
        A waiter future, what's result is set when connection is made, aka handshake is completed, or if when the
        connection is lost, depending which happens first.
        
        After the future's result or exception is set, the attribute is set as `None`.
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _in_handshake : `bool`
        Whether the ssl transport is in handshaking.
    _in_shutdown : `bool`
        Whether the ssl protocol is shut or shutting shutting down.
    _session_established : `bool`
        Whether the session is established. Is set after handshake and is set back to `False` when the connection is
        lost.
    _ssl_context : `ssl.SSLContext`
        The connection's ssl type.
    _write_backlog : `deque` of `tuple` (`bytes-like`, `int`)
        Ensured data queued up to be written. Each element contains a tuple of the data to write and an offset till
        write from.
    app_protocol : `Any`
        Asynchronous protocol implementation.
    app_transport : ``_SSLProtocolTransport``
        Asynchronous ssl protocol's read-write  transport implementation.
    loop : ``EventThread``
        The respective event loop of the ssl protocol.
    server_hostname : `None` or `str`
        The ssl protocol's server hostname if applicable.
    server_side : `bool`
        Whether the ssl protocol is server side.
    ssl_pipe : `None` or ``_SSLPipe``
        Ssl pipe set meanwhile the protocol is connected to feed the ssl data to.
    transport : `None` or `Any`
        Asynchronous transport implementation.
    """
    __slots__ = ('_call_connection_made', '_connection_made_waiter', '_extra', '_handshake_start_time', '_in_handshake',
        '_in_shutdown', '_session_established', '_ssl_context', '_write_backlog', 'app_protocol', 'app_transport',
        'loop', 'server_hostname', 'server_side', 'ssl_pipe', 'transport',)
    
    def __new__(cls, loop, app_protocol, ssl_context, connection_made_waiter, server_side=False, server_hostname=None,
            call_connection_made=True):
        """
        Creates a new ``SSLProtocol`` instance.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop of the protocol.
        app_protocol : `Any`
            Asynchronous protocol implementation.
        ssl_context : `None` or `ssl.SSLContext`
            The connection ssl type. If SSl was given as `True` when creating a connection at this point we get it as
            `None`, so we create a default ssl context.
            
            Note, that if the connection is server side, a valid `ssl_context` should be given.
        connection_made_waiter : `None` or ``Future``
            A waiter future, what's result is set when connection is made, aka handshake is completed, or if when the
            connection is lost, depending which happens first.
            
            After the future's result or exception is set, the attribute is set as `None`.
        server_side
            Whether the ssl protocol is server side.
        server_hostname
            The ssl protocol's server hostname if applicable.
            
            If we are the `server_side`, then this parameter is forced to `None` (wont raise).
        call_connection_made
            Whether the the ``.app_protocol``'s `.connection_made` should be called when handshake is completed.
        
        Raises
        ------
        ValueError
            If the protocol is server side, but `ssl_context` is not given.
        """
        if ssl_context is None:
            if server_side:
                raise ValueError('Server side SSL needs a valid `ssl.SSLContext`.')
            
            ssl_context = ssl.create_default_context()
            if (server_hostname is None) or (not server_hostname):
                ssl_context.check_hostname = False
        
        if server_side:
            server_hostname = None
        
        self = object.__new__(cls)
        
        self.server_side = server_side
        self.server_hostname = server_hostname
        self._ssl_context = ssl_context
        self._extra = {'sslcontext': ssl_context}
        self._write_backlog = deque()
        self._connection_made_waiter = connection_made_waiter
        self.loop = loop
        self.app_protocol = app_protocol
        self.app_transport = _SSLProtocolTransport(self, app_protocol)
        self.ssl_pipe = None
        self._session_established = False
        self._in_handshake = False
        self._in_shutdown = False
        self.transport = None
        self._call_connection_made = call_connection_made
        
        return self
    
    def _wake_up_connection_made_waiter(self, exception=None):
        """
        Wakes up the ssl protocol's ``._connection_made_waiter`` if applicable.
        
        Called when connection is made closed or lost, depending, which happens first. If connection is made or closed,
        then the connection made waiter's result is set as `None`, if the connection is lost, then the respective
        exception is thrown into it if any.
        
        Parameters
        ----------
        exception : `None` or ``BaseException``, Optional
            Exception to throw into ``._connection_made_waiter`` if any.
        """
        connection_made_waiter = self._connection_made_waiter
        if (connection_made_waiter is not None):
            self._connection_made_waiter = None
            if connection_made_waiter.pending():
                if exception is None:
                    connection_made_waiter.set_result(None)
                else:
                    connection_made_waiter.set_exception(exception)
    
    def connection_made(self, transport):
        """
        Called when the connection is made.
        
        Sets the ``SSLProtocol``'s ``.ssl_pipe`` and starts the ssl handshake.
        
        Parameters
        ----------
        transport : `Any`
            Asynchronous transport implementation, what calls the protocol's ``.data_received`` when data is
            received.
        """
        self.transport = transport
        self.ssl_pipe = _SSLPipe(self._ssl_context, self.server_side, self.server_hostname)
        
        self._handshake_start_time = None
        self._in_handshake = True
        # `(b'', 1)` is a special value in ``._process_write_backlog`` to do the SSL handshake
        self._write_backlog.append((b'', 1))
        self.loop.call_soon(self._process_write_backlog)
    
    def connection_lost(self, exception):
        """
        Called when the connection is lost or closed.
        
        Parameters
        ----------
        exception : `None` or `BaseException` instance
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        if self._session_established:
            self._session_established = False
            app_protocol = self.app_protocol
            self.loop.call_soon(app_protocol.__class__.connection_lost, app_protocol, exception)
        
        self.transport = None
        self.app_transport = None
        self._wake_up_connection_made_waiter(exception)
    
    def pause_writing(self):
        """
        Called when the transport's buffer goes over the high-water mark.
        
        ``.pause_writing`` is called when the buffer goes over the high-water mark, and eventually
        ``.resume_writing`` is called when the buffer size reaches the low-water mark.
        """
        self.app_protocol.pause_writing()
    
    def resume_writing(self):
        """
        Called when the transport's buffer drains below the low-water mark.
        
        See ``.pause_writing`` for details.
        """
        self.app_protocol.resume_writing()
    
    def data_received(self, data):
        """
        Called when some data is received.
        
        Parameters
        ----------
        data : `bytes`
            The received data.
        """
        try:
            ssl_data, application_data = self.ssl_pipe.feed_ssl_data(data)
        except ssl.SSLError:
            self._abort()
            return
        
        for chunk in ssl_data:
            self.transport.write(chunk)
        
        for chunk in application_data:
            if chunk:
                self.app_protocol.data_received(chunk)
                continue
            
            self._start_shutdown()
            break
    
    def eof_received(self):
        """
        Calling ``.connection_lost`` without exception causes eof.
        
        Marks the protocol as it is at eof.
        
        If the protocol's ``._connection_made_waiter`` is not yet waken up, then raises `ConnectionResetError` into it.
        
        Returns
        -------
        transport_closes : `bool`
            Returns `False` if the transport will close itself. If it returns `True`, then closing the transport is up
            to the protocol.
            
            Always returns `False`.
        """
        try:
            self._wake_up_connection_made_waiter(ConnectionResetError)
            if not self._in_handshake:
                # has no effect whatever it returns when we use ssl
                self.app_protocol.eof_received()
        finally:
            self.transport.close()
        
        return False
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        try:
            extra_info = self._extra[name]
        except KeyError:
            transport = self.transport
            if transport is not None:
                extra_info = transport.get_extra_info(name, default)
            else:
                extra_info = default
        
        return extra_info
    
    def _start_shutdown(self):
        """
        Starts the shutdown process of the ``SSLProtocol`` by writing empty bytes into itself.
        
        If the protocol is already in shutdown, does nothing.
        """
        if not self._in_shutdown:
            self._in_shutdown = True
            self._write_application_data(b'')
    
    def _write_application_data(self, data):
        """
        Writes data to the ``SSLProtocol`` to be sent.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to write.
        """
        self._write_backlog.append((data, 0))
        self._process_write_backlog()
    
    def _on_handshake_complete(self, handshake_exception):
        """
        Called when ssl handshake is completed.
        
        Parameters
        ----------
        handshake_exception : `None` or `BaseException` instance
        
        """
        self._in_handshake = False
        ssl_object = self.ssl_pipe.ssl_object
        
        if (handshake_exception is None):
            try:
                peer_certification = ssl_object.getpeercert()
            except BaseException as err:
                handshake_exception = err
            else:
                handshake_exception = None
        
        if (handshake_exception is not None):
            self.transport.close()
            self._wake_up_connection_made_waiter(handshake_exception)
            raise
        
        # Add extra info that becomes available after handshake.
        extra = self._extra
        extra['peercert'] = peer_certification
        extra['cipher'] = ssl_object.cipher()
        extra['compression']= ssl_object.compression()
        extra['ssl_object'] = ssl_object
        
        if self._call_connection_made:
            self.app_protocol.connection_made(self.app_transport)
        
        self._wake_up_connection_made_waiter()
        self._session_established = True
        
        # Don's call it immediately, other tasks might be scheduled already.
        self.loop.call_soon(self._process_write_backlog)

    def _process_write_backlog(self):
        """
        Try to make progress on the write backlog.
        
        Feeds data to ``.ssl_pipe`` till it is full.
        """
        transport = self.transport
        if transport is None:
            return
        
        write_backlog = self._write_backlog
        try:
            for _ in range(len(write_backlog)):
                data, offset = write_backlog[0]
                if data:
                    ssl_data, offset = self.ssl_pipe.feed_application_data(data, offset)
                else:
                    # If data is given as empty bytes means we either have handshake complete or there is no more data
                    # to write
                    if offset:
                        ssl_data = self.ssl_pipe.do_handshake(self._on_handshake_complete)
                    else:
                        ssl_data = self.ssl_pipe.shutdown(self._finalize)
                    
                    offset = 1
                
                for chunk in ssl_data:
                    transport.write(chunk)
                
                if offset < len(data):
                    write_backlog[0] = (data, offset)
                    # A short write means that a write is blocked on a read, we need to enable reading if it is paused!
                    assert self.ssl_pipe.need_ssl_data
                    if transport._paused:
                        transport.resume_reading()
                    
                    break
                
                # An entire chunk from the backlog was processed. We can delete it and reduce the outstanding buffer
                # size.
                del self._write_backlog[0]
        except BaseException as err:
            if self._in_handshake:
                self._on_handshake_complete(err)
            else:
                self._fatal_error(err, 'Fatal error on SSL transport.')
            if not isinstance(err, Exception):
                # BaseException
                raise
    
    def _fatal_error(self, exception, message='Fatal error on transport.'):
        """
        If a fatal error occurs on the protocol renders it's traceback and closes it's transport.
        
        Parameters
        ----------
        exception : `BaseException`
            The occurred exception.
        message : `str`, Optional
            Additional error message to render.
        """
        if not isinstance(exception, (BrokenPipeError, ConnectionResetError, ConnectionAbortedError)):
            self.loop.render_exc_async(exception, [
                message,
                ' exception occurred\n',
                repr(self),
                '\n',
                    ])
        
        transport = self.transport
        if (transport is not None):
            transport._force_close(exception)

    def _finalize(self):
        """
        CLoses the ``SSLProtocol``'s transport.
        
        Called after shutdown or abortion.
        """
        transport = self.transport
        if (transport is not None):
            transport.close()

    def _abort(self):
        """
        Closes the protocol by trying to abort it's transport immediately.
        """
        transport = self.transport
        if (transport is not None):
            try:
                transport.abort()
            finally:
                self._finalize()


class _SelectorSocketTransport:
    """
    Socket transport for selector based event loops.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether ``.write_eof`` was called.
    _connection_lost : `bool`
        Set as `True`, when ``._call_connection_lost`` is scheduled.
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _high_water : `int`
        The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
    _low_water : `int`
        The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to `16384`.
    _paused : `bool`
        Whether the transport's reading is paused by the protocol.
    _socket_fd : `int`
        The transport's socket's file descriptor identifier.
    closing : `bool`
        Whether the transport ic closing.
    loop : ``EventThread``
        The event loop to what the transport is bound to.
    protocol : `None`, ``SSLProtocol, ``ReadProtocolBase`` or `Any`
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    protocol_paused : `bool`
        Whether ``.protocol`` is paused writing.
    server : `None` or ``Server``
        If the transport is server side, it's server is set as this attribute.
    socket : `socket.socket`
        The socket used by the transport.
    """
    __slots__ = ('_at_eof', '_buffer', '_connection_lost', '_extra', '_high_water', '_low_water', '_paused',
        '_socket_fd', 'closing',  'loop', 'protocol', 'protocol_paused', 'server', 'socket')
    
    def __init__(self, loop, socket, protocol, waiter=None, extra=None, server=None):
        """
        Creates a ``_SelectorSocketTransport`` instance with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the transport is bound to.
        socket : `socket.socket`
            The socket used by the transport.
        protocol : ``SSLProtocol, ``ReadProtocolBase`` or `Any`
            Asynchronous protocol implementation used by the transport.
        waiter : `None` or ``Future`, Optional
            Waiter, what's result is set, when the transport connected. Defaults to `None`.
        extra : `None` or `dict` of (`str`, `Any`) items, Optional
            Optional transport information.
        server : `None` or ``Server``, Optional
            If the transport is server side, it's server is set as this attribute. Defaults to `None`.
        """
        if extra is None:
            extra = {}
        
        self._extra = extra
        self.loop = loop
        self.protocol_paused = False
        self._set_write_buffer_limits()
        
        extra['socket'] = socket
        
        try:
            socket_name = socket.getsockname()
        except OSError:
            socket_name = None
        extra['sockname'] = socket_name
        
        if 'peername' not in extra:
            try:
                peer_name = socket.getpeername()
            except module_socket.error:
                peer_name = None
            extra['peername'] = peer_name
        
        self.socket = socket
        self._socket_fd = socket.fileno()
        
        self.protocol = protocol
        
        self.server = server
        self._buffer = bytearray()
        self._connection_lost = False  # Set when call to connection_lost scheduled.
        self.closing = False  # Set when close() called.
        if server is not None:
            server._attach()
        
        self._at_eof = False
        self._paused = False
        
        # Disable the Nagle algorithm -- small writes will be sent without waiting for the TCP ACK.  This generally
        # decreases the latency (in some cases significantly.)
        _set_nodelay(socket)
        
        loop.call_soon(protocol.connection_made, self)
        
        #only start reading when connection_made() has been called
        loop.call_soon(loop.add_reader, self._socket_fd, self._read_ready)
        if (waiter is not None):
            # only wake up the waiter when connection_made() has been called
            loop.call_soon(Future.set_result_if_pending, waiter, None)
    
    def __repr__(self):
        """Returns the selector socket transport's representation."""
        result = [
            '<',
            self.__class__.__name__,
                ]
        
        if self.socket is None:
            result.append(' closed')
        elif self.closing:
            result.append(' closing')
        
        result.append(' fd=')
        result.append(repr(self._socket_fd))
        
        loop = self.loop
        #is the transport open?
        if (loop is not None) and loop.running:
        
            try:
                key = loop.selector.get_key(self._socket_fd)
            except KeyError:
                polling = 0
            else:
                polling = key.events&selectors.EVENT_READ
            
            result.append(' read=')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            result.append(state)

            try:
                key = loop.selector.get_key(self._socket_fd)
            except KeyError:
                polling = 0
            else:
                polling = key.events&selectors.EVENT_WRITE

            result.append(' write=<')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            result.append(state)

            result.append(', buffer_size=')
            
            buffer_size = self.get_write_buffer_size()
            result.append(str(buffer_size))
            result.append('>')
        
        result.append('>')
        
        return ''.join(result)
    
    def __del__(self):
        """
        Closes the ``._SelectorSocketTransport``'s ``.socket`` if not yet closed.
        """
        socket = self.socket
        if socket is not None:
            socket.close()
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self._extra.get(name, default)
    
    def get_protocol(self):
        """
        Gets the transport's actual protocol.
        
        Returns
        -------
        protocol : `None`, ``SubprocessWritePipeProtocol` or  `Any`
            Asynchronous protocol implementation.
        """
        return self.protocol
    
    def set_protocol(self, protocol):
        """
        Sets a new protocol to the transport.
        
        Parameters
        ----------
        protocol : `Any`
            Asynchronous protocol implementation.
        """
        self.protocol = protocol
    
    def is_closing(self):
        """
        Returns whether the transport is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self.closing
    
    def close(self):
        """
        Starts the shutdown process of the transport.
        
        If the transport is already closing, does nothing.
        """
        if self.closing:
            return
        
        self.closing = True
        self.loop.remove_reader(self._socket_fd)
        if not self._buffer:
            self._connection_lost = True
            self.loop.remove_writer(self._socket_fd)
            self.loop.call_soon(self._call_connection_lost, None)
    
    def abort(self):
        """
        Closes the transport immediately.
        
        The buffered data will be lost.
        """
        self._force_close(None)
    
    def write(self, data):
        """
        Write the given `data` to the transport.
        
        Do not blocks, but queues up the data instead to be sent as can.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to send.
        
        Raises
        ------
        TypeError
            If `data` was not given as `bytes-like`.
        RuntimeError : `bool`
            If ``.write_eof`` was already called.
        """
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(f'data parameter must should be `bytes-like`, got {data.__class__.__name__}.')
        if self._at_eof:
            raise RuntimeError('Cannot call `.write` after `.write_eof`')
        
        if not data:
            return
        
        if self._connection_lost:
            return
        
        if not self._buffer:
            # Optimization: try to send now.
            try:
                n = self.socket.send(data)
            except (BlockingIOError, InterruptedError):
                pass
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on socket transport')
                return
            else:
                data = data[n:]
                if not data:
                    return
            # Not all was written; register write handler.
            self.loop.add_writer(self._socket_fd, self._write_ready)
        
        # Add it to the buffer.
        self._buffer.extend(data)
        self._maybe_pause_protocol()
    
    def writelines(self, lines):
        """
        Writes the given lines to the transport's socket.
        
        Parameters
        ----------
        lines : `iterable` of `bytes-like`
            The lines to write.
        
        Raises
        ------
        RuntimeError
            Protocol has no attached transport.
        """
        self.write(b''.join(lines))
    
    def write_eof(self):
        """
        Writes eof to the subprocess pipe's transport's protocol if applicable.
        
        By default ``SubprocessStreamWriter``'s transport is ``UnixWritePipeTransport``, what will call connection lost
        as well when the write buffer is empty.
        """
        if self._at_eof:
            return
        
        self._at_eof = True
        
        if not self._buffer:
            self.socket.shutdown(module_socket.SHUT_WR)
    
    def can_write_eof(self):
        """
        Return whether the transport supports ``.write_eof``.
        
        Returns
        -------
        can_write_eof : `bool`
            ``_SelectorSocketTransport`` instances always return `True`.
        """
        return True
    
    def get_write_buffer_size(self):
        """
        Return the current size of the write buffer.
        
        Returns
        -------
        get_write_buffer_size : `int`
        """
        return len(self._buffer)
    
    def get_write_buffer_limits(self):
        """
        Returns the low and the high water of the transport.
        
        Returns
        -------
        low_water : `int`
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.
        high_water : `int`
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to
            `65536`.
        """
        return self._low_water, self._high_water
    
    def set_write_buffer_limits(self, high=None, low=None):
        """
        Set the high- and low-water limits for write flow control.
        
        These two values control when to call the protocol's ``.pause_writing`` and ``.resume_writing`` methods. If
        specified, the low-water limit must be less than or equal to the high-water limit. Neither value can be
        negative. The defaults are implementation-specific. If only the high-water limit is given, the low-water limit
        defaults to an implementation-specific value less than or equal to the high-water limit. Setting high to zero
        forces low to zero as well, and causes ``.pause_writing`` to be called whenever the buffer becomes non-empty.
        Setting low to zero causes ``.resume_writing`` to be called only once the buffer is empty. Use of zero for
        either limit is generally sub-optimal as it reduces opportunities for doing I/O and computation concurrently.
        
        Parameters
        ----------
        high : `None` or `int`, Optional
            High limit to stop reading if reached.
        low : `None` or `int`, Optional
            Low limit to start reading if reached.
        
        Raises
        ------
        ValueError
            If `low` is lower than `0` or if `low` is higher than `high`.
        """
        self._set_write_buffer_limits(high=high, low=low)
        self._maybe_pause_protocol()
    
    def pause_reading(self):
        """
        Pauses the receiving end.
        
        No data will be passed to the respective protocol's ``.data_received`` method until ``.resume_reading`` is
        called.
        
        Returns
        -------
        reading_paused : `bool`
            Whether reading was paused.
        """
        if self.closing or self._paused:
            return False
        
        self._paused = True
        self.loop.remove_reader(self._socket_fd)
        return True
    
    def resume_reading(self):
        """
        Resumes the receiving end.
        
        Data received will once again be passed to the respective protocol's ``.data_received`` method.
        
        Returns
        -------
        reading_resumed : `bool`
            Whether reading was resumed.
        """
        if not self._paused:
            return False
        
        self._paused = False
        if not self.closing:
            self.loop.add_reader(self._socket_fd, self._read_ready)
        
        return True
    
    def _maybe_pause_protocol(self):
        """
        Called after data was ensured to be written into the socket to check whether it's protocol should be paused.
        """
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        
        if self.protocol_paused:
            return
        
        self.protocol_paused = True
        try:
            self.protocol.pause_writing()
        except BaseException as err:
            self.loop.render_exc_async(err, [
                'Exception occurred at:\n',
                repr(self),
                '._maybe_pause_protocol\n',
                    ])
    
    def _maybe_resume_protocol(self):
        """
        Called after successful writing to the socket to check whether the protocol should be resumed.
        """
        if (not self.protocol_paused) or (self.get_write_buffer_size() > self._low_water):
            return
        self.protocol_paused = False
        try:
            self.protocol.resume_writing()
        except BaseException as err:
            self.loop.render_exc_async(err, [
                'Exception occurred at:\n',
                repr(self),
                '._maybe_resume_protocol\n',
            ])
    
    def _set_write_buffer_limits(self, high=None, low=None):
        """
        Sets the write buffer limits of the transport.
        
        Parameters
        ----------
        low : None` or `int`, Optional
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
        high : `None` or `int`, Optional
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.

        Raises
        ------
        ValueError
            If `high` is lower than `low` or if `low` is lower than `0`.
        """
        if high is None:
            if low is None:
                high = 65536
                low = 16384
            else:
                high = low<<2
        else:
            if low is None:
                low = high>>2
        
        if low < 0 or high < low:
            raise ValueError(f'High water must be greater or equal than low, what must be greater than equal than `0`, '
                f'got high={high!r}; low={low!r}.')
        
        self._high_water = high
        self._low_water = low
    
    def _fatal_error(self, exception, message='Fatal error on transport'):
        """
        If a fatal error occurs on the transport, renders its traceback and closes itself.
        
        Parameters
        ----------
        exception : `BaseException`
            The occurred exception.
        message : `str`, Optional
            Additional error message to render.
        """
        if not isinstance(exception, (BrokenPipeError, ConnectionResetError, ConnectionAbortedError)):
            self.loop.render_exc_async(exception, [
                message,
                ' exception occurred\n',
                repr(self),
                '\n',
            ])
        
        self._force_close(exception)
    
    def _force_close(self, exception):
        """
        Closes the transport immediately.
        
        The buffered data will be lost.
        
        Parameters
        ----------
        exception : `None` or ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        if self._connection_lost:
            return
        
        buffer = self._buffer
        if buffer:
            buffer.clear()
            self.loop.remove_writer(self._socket_fd)
        
        if not self.closing:
            self.closing = True
            self.loop.remove_reader(self._socket_fd)
        
        self._connection_lost = True
        self.loop.call_soon(self._call_connection_lost, exception)
    
    def _call_connection_lost(self, exception):
        """
        Calls the transport's connection lost method-
        
        Parameters
        ----------
        exception : `None` or ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        try:
            protocol = self.protocol
            if (protocol is not None):
                protocol.connection_lost(exception)
        finally:
            self.socket.close()
            self.socket = None
            self.protocol = None
            self.loop = None
            server = self.server
            if (server is not None):
                server._detach()
                self.server = None
    
    def _read_ready(self):
        """
        Added as a read callback on the respective event loop to be called when the data is received on the pipe.
        
        If this happens, since it is a write only pipe, means it should be closed, so we do like that.
        """
        if self._connection_lost:
            return
        try:
            data = self.socket.recv(MAX_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        except BaseException as err:
            self._fatal_error(err, 'Fatal read error on socket transport')
        else:
            if data:
                self.protocol.data_received(data)
            elif self.protocol.eof_received():
                # We're keeping the connection open so the
                # protocol can write more, but we still can't
                # receive more, so remove the reader callback.
                self.loop.remove_reader(self._socket_fd)
            else:
                self.close()

    def _write_ready(self):
        """
        Added as a write callback on the respective event loop when the transport has unsent data. Called when the
        respective socket becomes writable.
        """
        if self._connection_lost:
            return
        
        try:
            n = self.socket.send(self._buffer)
        except (BlockingIOError, InterruptedError):
            pass
        except BaseException as err:
            self.loop.remove_writer(self._socket_fd)
            self._buffer.clear()
            self._fatal_error(err, 'Fatal write error on socket transport')
        else:
            if n:
                del self._buffer[:n]
            
            self._maybe_resume_protocol()  # May append to buffer.
            if not self._buffer:
                self.loop.remove_writer(self._socket_fd)
                
                if self.closing:
                    self._call_connection_lost(None)
                
                elif self._at_eof:
                    self.socket.shutdown(module_socket.SHUT_WR)


class _SelectorDatagramTransport:
    """
    Asynchronous transport implementation for datagram sockets.
    
    _buffer : `deque` of `tuple` (`bytes`, `tuple` (`str`, `int`))
        A queue of received data-s, each in `data-address` tuples.
    _connection_lost : `bool`
        Set as `True`, when ``._call_connection_lost`` is scheduled.
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _high_water : `int`
        The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
    _low_water : `int`
        The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to `16384`.
    _paused : `bool`
        Whether the transport's reading is paused by the protocol.
    _socket_fd : `int`
        The transport's socket's file descriptor identifier.
    address : `None` or (`str`, `int`)
        The last address, where the transport sent data. Defaults to `None`. The send target address should not differ
        from the last, where the transport sent data.
    closing : `bool`
        Whether the transport ic closing.
    loop : ``EventThread``
        The event loop to what the transport is bound to.
    protocol : `None`, ``DatagramAddressedReadProtocol``, ``DatagramMergerReadProtocol`` or `Any`
        Asynchronous protocol implementation used by the transport.
        
        After closing the transport is set to `None`.
    protocol_paused : `bool`
        Whether ``.protocol`` is paused writing.
    socket : `socket.socket`
        The socket used by the transport.
    """
    __slots__ = ('_buffer', '_connection_lost', '_extra', '_high_water', '_low_water', '_socket_fd', 'address',
        'closing', 'loop', 'protocol', 'protocol_paused', 'socket')
    
    def __init__(self, loop, socket, protocol, address=None, waiter=None, extra=None):
        """
        Creates a ``_SelectorDatagramTransport`` instance with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the transport is bound to.
        socket : `socket.socket`
            The socket used by the transport.
        protocol : ``SSLProtocol, ``ReadProtocolBase`` or `Any`
            Asynchronous protocol implementation used by the transport.
        address : `None` or `tuple` (`str`, `int`), Optional
            The last address, where the transport sent data. Defaults to `None`. The send target address should not
            differ from the last, where the transport sent data.
        waiter : `None` or ``Future`, Optional
            Waiter, what's result is set, when the transport connected. Defaults to `None`.
        extra : `None` or `dict` of (`str`, `Any`) items, Optional
            Optional transport information.
        """
        if extra is None:
            extra = {}
        
        self._extra = extra
        
        self.loop = loop
        self.protocol_paused = False
        self._set_write_buffer_limits()
        
        extra['socket'] = socket
        
        try:
            socket_name = socket.getsockname()
        except OSError:
            socket_name = None
        extra['sockname'] = socket_name
        
        if 'peername' not in extra:
            try:
                peer_name = socket.getpeername()
            except module_socket.error:
                peer_name = None
            
            extra['peername'] = peer_name
        
        self.socket = socket
        self._socket_fd = socket.fileno()
        
        self.protocol = protocol
        
        self._buffer = deque()
        self._connection_lost = 0 # Set when call to connection_lost scheduled.
        self.closing = False # Set when close() called.
        
        self.address = address
        loop.call_soon(self.protocol.connection_made, self)
        # only start reading when connection_made() has been called
        loop.call_soon(self._add_reader)
        
        if waiter is not None:
            # only wake up the waiter when connection_made() has been called
            loop.call_soon(Future.set_result_if_pending, waiter, None)
    
    def __repr__(self):
        """Returns the transport's representation."""
        result = [
            '<',
            self.__class__.__name__,
                ]
        
        if self.socket is None:
            result.append(' closed')
        elif self.closing:
            result.append(' closing')
        
        result.append(' fd=')
        result.append(repr(self._socket_fd))
        
        loop = self.loop
        #is the transport open?
        if (loop is not None) and loop.running:
            try:
                key = loop.selector.get_key(self._socket_fd)
            except KeyError:
                polling = 0
            else:
                polling = key.events&selectors.EVENT_READ
            
            result.append(' read=')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            result.append(state)
            
            try:
                key = loop.selector.get_key(self._socket_fd)
            except KeyError:
                polling = 0
            else:
                polling = key.events&selectors.EVENT_WRITE
            
            result.append(' write=<')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            result.append(state)
            
            result.append(', buffer_size=')
            
            buffer_size = self.get_write_buffer_size()
            result.append(str(buffer_size))
            result.append('>')
        
        result.append('>')
        
        return ''.join(result)
    
    def __del__(self):
        """
        Closes the ``._SelectorSocketTransport``'s ``.socket`` if not yet closed.
        """
        socket = self.socket
        if (socket is not None):
            socket.close()
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self._extra.get(name, default)
    
    def get_protocol(self):
        """
        Gets the transport's actual protocol.
        
        Returns
        -------
        protocol : `None`, ``SubprocessWritePipeProtocol` or  `Any`
            Asynchronous protocol implementation.
        """
        return self.protocol
    
    def set_protocol(self, protocol):
        """
        Sets a new protocol to the transport.
        
        Parameters
        ----------
        protocol : `Any`
            Asynchronous protocol implementation.
        """
        self.protocol = protocol
    
    def is_closing(self):
        """
        Returns whether the transport is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self.closing
    
    def close(self):
        """
        Starts the shutdown process of the transport.
        
        If the transport is already closing, does nothing.
        """
        if self.closing:
            return
        
        self.closing = True
        self.loop.remove_reader(self._socket_fd)
        if not self._buffer:
            self._connection_lost += 1
            self.loop.remove_writer(self._socket_fd)
            self.loop.call_soon(self._call_connection_lost, None)
    
    def pause_reading(self):
        """
        Pauses the receiving end.
        
        ``_SelectorDatagramTransport``-s do not support this operation.
        
        Returns
        -------
        reading_paused : `bool`
            Whether reading was paused. Always returns `False`.
        """
        return False
    
    def resume_reading(self):
        """
        Resumes the receiving end.
        
        ``_SelectorDatagramTransport``-s do not support this operation.
        
        Returns
        -------
        reading_resume : `bool`
            Whether reading was resumed. Always returns `False`.
        """
        return False
    
    def write(self, data):
        """
        Write the given data to transport.
        
        Parameters
        ----------
        data : `bytes-like`
            The bytes data to be sent.
        """
        return
    
    def writelines(self, lines):
        """
        Writes the given lines to the protocol's transport. 
        
        ``_SelectorDatagramTransport`` do not support writing, use ``.send_to`` instead.
        
        Parameters
        ----------
        lines : `iterable` of `bytes-like`
            The lines to write.
        """
        pass
        # Do not bother to join:
        # data = b''.join(lines)
        # self.write(data)
    
    def write_eof(self):
        """
        Writes eof to the transport's protocol if applicable.
        
        ``_SelectorDatagramTransport`` instances do not support this operation.
        """
        pass
    
    def can_write_eof(self):
        """
        Return whether the transport supports ``.write_eof``.
        
        Returns
        -------
        can_write_eof : `bool`
            ``_SelectorDatagramTransport`` do not support this operation, so this method will always return `False`.
        """
        return False
    
    def _maybe_pause_protocol(self):
        """
        Called after data was ensured to be written into the socket to check whether it's protocol should be paused.
        """
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        
        if self.protocol_paused:
            return
            
        self.protocol_paused = True
        try:
            self.protocol.pause_writing()
        except BaseException as err:
            self.loop.render_exc_async(err, [
                'Exception occurred at:\n',
                repr(self),
                '._maybe_pause_protocol\n',
                    ])
    
    def _maybe_resume_protocol(self):
        """
        Called after successful writing to the socket to check whether the protocol should be resumed.
        """
        if (not self.protocol_paused) or (self.get_write_buffer_size() > self._low_water):
            return
        
        self.protocol_paused = False
        try:
            self.protocol.resume_writing()
        except BaseException as err:
            self.loop.render_exc_async(err, [
                'Exception occurred at:\n',
                repr(self),
                '._maybe_resume_protocol\n',
                    ])
    
    def get_write_buffer_limits(self):
        """
        Returns the low and the high water of the transport.
        
        Returns
        -------
        low_water : `int`
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
        high_water : `int`
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.
        """
        return self._low_water, self._high_water
    
    def _set_write_buffer_limits(self, high=None, low=None):
        """
        Sets the write buffer limits of the transport.
        
        Parameters
        ----------
        low : None` or `int`, Optional
            The ``.protocol`` is paused writing when the buffer size passes the high water mark. Defaults to `65536`.
        high : `None` or `int`, Optional
            The ``.protocol`` is resumed writing when the buffer size goes under the low water mark. Defaults to
            `16384`.

        Raises
        ------
        ValueError
            If `high` is lower than `low` or if `low` is lower than `0`.
        """
        if high is None:
            if low is None:
                high = 65536
            else:
                high = low<<2
        
        if low is None:
            low = high>>2
        
        if not high >= low >= 0:
            raise ValueError(f'`high` ({high}) must be `>= low` ({low}) must be `>= 0`.')
        
        self._high_water = high
        self._low_water = low
    
    def set_write_buffer_limits(self, high=None, low=None):
        """
        Set the high- and low-water limits for write flow control.
        
        These two values control when to call the protocol's ``.pause_writing`` and ``.resume_writing`` methods. If
        specified, the low-water limit must be less than or equal to the high-water limit. Neither value can be
        negative. The defaults are implementation-specific. If only the high-water limit is given, the low-water limit
        defaults to an implementation-specific value less than or equal to the high-water limit. Setting high to zero
        forces low to zero as well, and causes ``.pause_writing`` to be called whenever the buffer becomes non-empty.
        Setting low to zero causes ``.resume_writing`` to be called only once the buffer is empty. Use of zero for
        either limit is generally sub-optimal as it reduces opportunities for doing I/O and computation concurrently.
        
        Parameters
        ----------
        high : `None` or `int`, Optional
            High limit to stop reading if reached.
        low : `None` or `int`, Optional
            Low limit to start reading if reached.
        
        Raises
        ------
        ValueError
            If `low` is lower than `0` or if `low` is higher than `high`.
        """
        self._set_write_buffer_limits(high=high, low=low)
        self._maybe_pause_protocol()
    

    

    
    def abort(self):
        """
        Closes the transport immediately.
        
        The buffered data will be lost.
        """
        self._force_close(None)
    


    
    def _fatal_error(self, exception, message='Fatal error on transport'):
        """
        If a fatal error occurs on the transport, renders its traceback and closes itself.
        
        Parameters
        ----------
        exception : `BaseException`
            The occurred exception.
        message : `str`, Optional
            Additional error message to render.
        """
        if not isinstance(exception, OSError):
            self.loop.render_exc_async(exception, [
                message,
                ' exception occurred\n',
                repr(self),
                '\n',
                    ])
        
        self._force_close(exception)
    
    def _force_close(self, exception):
        """
        Closes the transport immediately.
        
        The buffered data will be lost.
        
        Parameters
        ----------
        exception : `None` or ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        if self._connection_lost:
            return
        
        if self._buffer:
            self._buffer.clear()
            self.loop.remove_writer(self._socket_fd)
        
        if not self.closing:
            self.closing = True
            self.loop.remove_reader(self._socket_fd)
        
        self._connection_lost += 1
        self.loop.call_soon(self._call_connection_lost, exception)
    
    def _call_connection_lost(self, exception):
        """
        Calls the transport's connection lost method-
        
        Parameters
        ----------
        exception : `None` or ``BaseException``
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        try:
            protocol = self.protocol
            if (protocol is not None):
                self.protocol.connection_lost(exception)
        finally:
            self.socket.close()
            self.socket = None
            self.protocol = None
            self.loop = None
    
    def _add_reader(self):
        """
        Call soon callback added by ``.__init__` to add reader to the event loop.
        """
        
        if self.closing:
            return
        
        self.loop.add_reader(self._socket_fd, self._read_ready)
    
    def get_write_buffer_size(self):
        """
        Return the current size of the write buffer.
        
        Returns
        -------
        get_write_buffer_size : `int`
        """
        size = 0
        for data, address in self._buffer:
            size += len(data)
        
        return size
    
    def _read_ready(self):
        """
        Added as a read callback on the respective event loop to be called when the data is received on the pipe.
        
        If this happens, since it is a write only pipe, means it should be closed, so we do like that.
        """
        if self._connection_lost:
            return
        try:
            data, address = self.socket.recvfrom(MAX_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as err:
            self.protocol.error_received(err)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as err:
            self._fatal_error(err, 'Fatal read error on datagram transport')
        else:
            self.protocol.datagram_received(data, address)
    
    def send_to(self, data, maybe_address=None):
        """
        Sends the given data to the target address. If target is already set, can be the parameter can be ignored.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to send.
        maybe_address : `None` or `tuple` (`str`, `int`), Optional
            The address to send the data to.
        
        Raises
        ------
        TypeError
            If `data` was not given as `bytes-like`.
        ValueError
            If `address` was given but it is different from the currently set one.
        """
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(f'data parameter must be a bytes-like object, not {type(data).__name__!r}')
        
        if not data:
            return
        
        address = self.address
        if (address is not None):
            if (maybe_address is not None) or (maybe_address != address):
                raise ValueError(f'Invalid address: must be `None` or `{address!r}`, got {maybe_address!r}.')
            maybe_address = address
        
        if self._connection_lost:
            return
        
        buffer = self._buffer
        if not buffer:
            # Attempt to send it right away first.
            try:
                if self._extra['peername'] is None:
                    self.socket.sendto(data, maybe_address)
                else:
                    self.socket.send(data)
            except (BlockingIOError, InterruptedError):
                self.loop.add_writer(self._socket_fd, self._send_to_ready)
            except OSError as err:
                self.protocol.error_received(err)
                return
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on datagram transport')
                return
            else:
                return
        
        # Ensure that what we buffer is immutable.
        buffer.append((bytes(data), maybe_address))
        self._maybe_pause_protocol()
    
    def _send_to_ready(self):
        """
        Added callback by `.send_to` to the respective event loop, when the socket is not ready for writing.
        
        This method tries to send the data again.
        """
        buffer = self._buffer
        while buffer:
            data, address = buffer.popleft()
            try:
                if self._extra['peername'] is None:
                    self.socket.sendto(data, address)
                else:
                    self.socket.send(data)
            except (BlockingIOError, InterruptedError):
                buffer.appendleft((data, address))  # Try again later.
                break
            except OSError as err:
                self.protocol.error_received(err)
                return
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on datagram transport')
                return
        
        self._maybe_resume_protocol() # May append to buffer.
        if not buffer:
            self.loop.remove_writer(self._socket_fd)
            if self.closing:
                self._call_connection_lost(None)



__all__ = ('ChannelStore', )

from ...backend.utils import copy_docs
from ...backend.export import export, include

from ..core import CHANNELS
from ..permission import Permission
from ..permission.permission import PERMISSION_NONE, PERMISSION_TEXT_AND_VOICE_DENY, PERMISSION_MASK_VIEW_CHANNEL

from ..preconverters import preconvert_snowflake, preconvert_str, preconvert_bool

from .channel_base import ChannelBase
from .channel_guild_base import ChannelGuildMainBase

parse_permission_overwrites = include('parse_permission_overwrites')


@export
class ChannelStore(ChannelGuildMainBase):
    """
    Represents a ``Guild`` store channel.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    _permission_cache : `None` or `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `0`
        The channel's parent's identifier.
    guild_id : `int`
        The channel's guild's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    nsfw : `bool`
        Whether the channel is marked as non safe for work.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `6`
        The default type, what the channel represents.
    INTERCHANGE : `tuple` of `int` = `(6,)`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    ORDER_GROUP : `int` = `0`
        An order group what defined which guild channel type comes after the other one.
    type : `int` = `6`
        The channel's Discord side type.
    """
    __slots__ = ('nsfw',) # guild channel store related
    
    DEFAULT_TYPE = 6
    ORDER_GROUP = 0
    INTERCHANGE = (6,)
    type = 6
    
    def __new__(cls, data, client, guild_id):
        """
        Creates a store channel from the channel data received from Discord. If the channel already exists and if
        it is partial, then updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        client : `None` or ``Client``
            The client, who received the channel's data, if any.
        guild_id : `int`
            The channel's guild's identifier.
        """
        channel_id = int(data['id'])
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = channel_id
            CHANNELS[channel_id] = self
        else:
            if self.clients:
                return self
        
        self._permission_cache = None
        self.name = data['name']
        self.nsfw = data.get('nsfw', False)
        
        self._init_parent_and_position(data, guild_id)
        self.permission_overwrites = parse_permission_overwrites(data)
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelBase._create_empty)
    def _create_empty(cls, channel_id, channel_type, guild_id):
        self = super(ChannelStore, cls)._create_empty(channel_id, channel_type, guild_id)
        
        self.nsfw = False
        
        return self
    
    
    @property
    @copy_docs(ChannelBase.display_name)
    def display_name(self):
        return self.name.lower()
    
    
    @copy_docs(ChannelBase._update_attributes)
    def _update_attributes(self, data):
        self._permission_cache = None
        self._set_parent_and_position(data)
        self.permission_overwrites = parse_permission_overwrites(data)
        
        self.name = data['name']
        self.nsfw = data.get('nsfw', False)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the channel and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
            
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +---------------------------+---------------------------------------------------+
        | Keys                      | Values                                            |
        +===========================+===================================================+
        | parent_id                 | `int`                                             |
        +---------------------------+---------------------------------------------------+
        | name                      | `str`                                             |
        +---------------------------+---------------------------------------------------+
        | nsfw                      | `bool`                                            |
        +---------------------------+---------------------------------------------------+
        | permission_overwrites     | `dict` of (`int`, ``PermissionOverwrite``) items  |
        +---------------------------+---------------------------------------------------+
        | position                  | `int`                                             |
        +---------------------------+---------------------------------------------------+
        """
        self._permission_cache = None
        old_attributes = {}
        
        name = data['name']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        nsfw = data.get('nsfw', False)
        if self.nsfw != nsfw:
            old_attributes['nsfw'] = self.nsfw
            self.nsfw = nsfw
        
        permission_overwrites = parse_permission_overwrites(data)
        if self.permission_overwrites != permission_overwrites:
            old_attributes['permission_overwrites'] = self.permission_overwrites
            self.permission_overwrites = permission_overwrites
        
        self._update_parent_and_position(data, old_attributes)
        
        return old_attributes
    
    
    @copy_docs(ChannelBase.permissions_for)
    def permissions_for(self, user):
        result = self._permissions_for(user)
        if not result&PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # store channels do not have text and voice related permissions
        result &= PERMISSION_TEXT_AND_VOICE_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelBase.permissions_for_roles)
    def permissions_for_roles(self, *roles):
        result = self._permissions_for_roles(roles)
        if not result&PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # store channels do not have text and voice related permissions
        result &= PERMISSION_TEXT_AND_VOICE_DENY
        return Permission(result)
    
    @classmethod
    def precreate(cls, channel_id, **kwargs):
        """
        Precreates the channel by creating a partial one with the given parameters. When the channel is loaded
        the precrated channel will be picked up. If an already existing channel would be precreated, returns that
        instead and updates that only, if that is a partial channel.
        
        Parameters
        ----------
        channel_id : `int` or `str`
            The channel's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the channel.
        
        Other Parameters
        ----------------
        name : `str`, Optional (Keyword only)
            The channel's ``.name``.
        nsfw : `int`, Optional (Keyword only)
            Whether the channel is marked as nsfw.
        
        Returns
        -------
        channel : ``ChannelStore``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        channel_id = preconvert_snowflake(channel_id, 'channel_id')
        
        if kwargs:
            processable = []
            
            try:
                name = kwargs.pop('name')
            except KeyError:
                pass
            else:
                name = preconvert_str(name, 'name', 2, 100)
                processable.append(('name', name))
            
            try:
                nsfw = kwargs.pop('nsfw')
            except KeyError:
                pass
            else:
                nsfw = preconvert_bool(nsfw, 'nsfw')
                processable.append(('nsfw', nsfw))
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}')
        
        else:
            processable = None
        
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = cls._create_empty(channel_id, cls.type, 0)
            CHANNELS[channel_id] = self
            
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self
